--/ globals
globalBorderX, globalBorderY = 6, 4 --docked
textPadding = 3
drawScale = 1
redraw = 1

mouseWheelAccum = 0

--
--align = {}
--align.left = 0x80
--align.top = 0x40
--align.vcenter = 0x20
--align.hcenter = 0x04
--align.bottom = 0x02
--align.right = 0x01
--)
--/ mac key's codes
  s_char = 0
  kbd = {}
  kbd.up = 30064
  kbd.down = 1685026670
  kbd.left = 1818584692
  kbd.right = 1919379572
  kbd.space = 32
  
--/ global gui coords
  guirect = {}
  guirect.tl = { x = 0, y = 0 } -- top left
  guirect.br = { x = 720, y = 1024 } -- bottom right
  guicur = {}
  guicur.usable = false
  guicur.pos = { x = 0, y = 0 }
  
fnt = {}
fnt.nm = "Avenir Next"
fnt.sz = { default = 34, delta = 6 }
gfx.setfont( 1, fnt.nm, fnt.sz.default )

clr = {}
clr.default = 0xa0a0a0ff
clr.selected = 0xc07020ff
clr.highlighted = 0xc07020ff
clr.black = 0x000000ff
clr.white = 0xffffffff
clr.darkgray = 0x404040ff
clr.gray = 0x808080ff
clr.lightgray = 0xc0c0c0ff
clr.red = 0x800000ff
clr.green = 0x008000ff
clr.blue = 0x000080ff
clr.txt = { default = clr.gray }


-- Convert a hex color 0xRrGgBb[Aa] to floats r,g,b[,a] 0..1 range
-- adopted to gfx.set()
local function hex2rgba( num )
  local snum = string.format( "%x", num )
  local red = string.sub( snum, 1, 2 )
  local green = string.sub( snum, 3, 4 )
  local blue = string.sub( snum, 5, 6 )
  local alpha = string.sub( snum, 7, 8 )
  local r = tonumber( red, 16 )
  local g = tonumber( green, 16 )
  local b = tonumber( blue, 16 )
  local a = tonumber( alpha, 16 )
  if r == nil then r = 0 else r = r / 255 end
  if g == nil then g = 0 else g = g / 255 end
  if b == nil then b = 0 else b = b / 255 end
  if a == nil then a = 0 else a = a / 255 end
  return r, g, b, a
end

local function SetColor( hex_color )
  local r, g, b, a = hex2rgba( hex_color )
  gfx.set( r, g, b, a )
end

----
----

function text( str, x, y, w, h, align, col, style, lineSpacing, vCenter )
  local ls = 11
  if lineSpacing ~= nil then ls = lineSpacing end
  local lineSpace = drawScale * ls
  SetColor( col or clr.text.default )
  local str = str or '---'
  local tmp, cnt = string.gsub( str, '<br>', '' )
  if cnt ~= 0 then -- line break
    local strOver = string.sub( str, ( string.find( str, '<br>' ) + 4 ), -1 )
    local thisLineY, nextLineY = nil, nil
    if vCenter ~= false then -- valign 2-line
      thisLineY, nextLineY = y - ( lineSpace / 2 ), y + ( lineSpace / 2 )
    else
      thisLineY, nextLineY = y, y + lineSpace
    end
    text( strOver, x, nextLineY, w, h, align, col, style, lineSpacing, vCenter )
    y = thisLineY
    str = string.sub( str, 0, ( string.find( str, '<br>' ) - 1 ) )
  end

  gfx.setfont( style or 1 )
  gfx.x, gfx.y = x, y
  gfx.drawstr( str, align or 0, x + ( w or 0 ), y + ( h or 0 ) )
end


---
---/ objects
---

function adoptChild( parent, o )
  if parent ~= nil then
    if parent.children == nil then parent.children = { o }
    else parent.children[ #parent.children + 1 ] = o end
    
    if parent.has_children_outside ~= 1 then
      if o.has_children_outside == 1 then
        parent.has_children_outside = 1
      else
        if o.x ~= nil and o.y ~= nil and parent.w ~= nil and parent.h ~= nil then
          if o.x < 0 or o.y < 0 or
            o.x + ( o.w or 1 ) > parent.w or
            o.y + ( o.h or 1 ) > parent.h then
            parent.has_children_outside = 1
          end
        end
      end
    end
  end
end

--/ base Element
Element = {}
function Element:new( parent, o )
local o = o or {}
  self.__index = self
  self.x, self.y = self.x or 0, self.y or 0
  adoptChild( parent, o )
  setmetatable( o, self )
  return o
end

function Element:doUpdateState()
  if self.updateState ~= nil then
    self:updateState()
  end
  if self.children ~= nil then
    for i, v in ipairs( self.children ) do
      if v.visible ~= false then v:doUpdateState() end
    end
  end
end

function Element:onSize()
  local crop = self.crop or false
  if self.children ~= nil and self.visible ~= false and self.crop ~= true then
    for i, v in ipairs( self.children ) do
      
      local bx, by = 0, 0
      if v.border == 'xy' or v.border == 'x' then bx = globalBorderX end
      if v.border == 'xy' or v.border == 'y' then bx = globalBorderY end
      
      local prevElX, prevElY, prevElW, prevElH = 0, 0, 0, 0
      if i > 1 then --there is a prev child
        prevElX, prevElY, prevElW, prevElH = 
          self.children[ i - 1 ].drawx, self.children[ i - 1 ].drawy,
          self.children[ i - 1 ].drawW, self.children[ i - 1 ].drawH
        if self.children[ i - 1 ].visible == false then prevElW = 0 end
      end
      
      if v.flexW ~= nil then
        if v.flexW == 'fill' then
          v.drawW = self.drawx + ( self.drawW or self.w ) -
            ( ( prevElX or self.drawx ) + ( prevElW or 0 ) ) + ( v.w or 0 )
        else
          v.drawW = ( v.w or 0 ) +
            ( ( self.drawW or self.w ) * ( v.flexW:sub( 1, -2 ) / 100 ) )
        end
      else v.drawW = v.w
      end
      
      if v.flexH ~= nil then v.drawH = ( v.h or 0 ) + 
        ( ( self.drawH or self.h ) * ( v.flexH:sub( 1, -2 ) / 100 ) )
      else v.drawH = v.h
      end
      
      v:position( self.drawx, self.drawy, self.drawW, self.drawH,
        prevElX, prevElY, prevElW, prevElH, bx, by )
      v:onSize() -- this child sizes its children
      
    end
  end
end

function Element:position(
  parentX, parentY, parentW, parentH,
  prevElX, prevElY, prevElW, prevElH,
  bx, by )

  if parentX == nil then parentX = 0 end
  if parentY == nil then parentY = 0 end
  if parentW == nil then parentW = 0 end
  if parentH == nil then parentH = 0 end
  self.drawx, self.drawy = parentX + self.x + bx, parentY + self.y + by
  
  if self.posx ~= nil then
    if self.posx == 'center' then
      self.drawx, self.drawy = ( parentW - self.drawW ) / 2, ( parentH - self.drawH ) / 2
    elseif self.posx == 'right' then
      self.drawx = parentW - self.w
    end
  end
  
  if self.flow ~= nil and self.flow ~= false then
    if prevElX == nil or prevElW == 0 then
      self.drawx, self.drawy =
        parentX + self.x + bx, parentY + self.y + by
      self.crop = false
      if ( parentX + parentW ) < ( self.drawx + self.drawW + bx ) then
        self.crop = true
      end
    else
      if ( prevElX + prevElW + bx + self.drawW + bx ) <= ( parentX + parentW ) then
        self.drawx, self.drawy = prevElX + prevElW + bx + self.x, prevEly + self.y
        self.crop = false
      elseif ( parentY + parentH ) > ( prevElY + prevElH + self.y + self.drawH + by ) then
        self.drawx, self.drawy = parentX + self.x, prevElY + prevElH + self.y
        self.crop = false
      else
        self.crop = true
      end
    end
  end
  self.crop = false
  
end

function Element:draw()
  gfx.set( 0, 0, 0, 1 )
  local crop = self.crop or false
  if self.visible ~= false and crop ~= true then
    local thisX = drawScale * ( self.drawx or self.x )
    local thisY = drawScale * ( self.drawy or self.y )
    local thisW = drawScale * ( self.drawW or self.w or 0 )
    local thisH = drawScale * ( self.drawH or self.h or 0 )
    local thisDrawW = drawScale * ( self.drawW or 0 )
    local thisDrawH = drawScale * ( self.drawH or 0 )
    
    local thisCol = self.drawColor or self.color or nil

    if thisCol == nil then thisCol = 0x180810ff end
    
    if thisCol ~= nil then
      SetColor( thisCol )
      gfx.rect( thisX, thisY, thisDrawW, thisDrawH )
    end
    
    if self.text ~= nil then
      if self.text.val ~= nil then
        self.text.str = self.text.val()
      end
      local txtScaleOffs = ''
      if drawScale == 2 then txtScaleOffs = 1 end
      local tx, tw = thisX + ( drawScale * textPadding ),
        thisW - 2 * ( drawScale * textPadding )
      text(
        self.text.str, tx, thisY, tw, thisDrawH,
        self.text.align, self.text.col,
        txtScaleOffs..(self.text.style or 1),
        self.text.lineSpacing, self.text.vCenter )
    end
    drawChildren( self.children )
  else -- crop
    
  end
end

function drawChildren( ch )
  if ch ~= nil then
    for i, v in ipairs( ch ) do
      v:draw()
    end
  end
end

function Element:kbd()
  if self.visible ~= false then
    kbdChildren( self.children )
  end
end
function kbdChildren( ch )
  if ch ~= nil then
    for i, v in ipairs( ch ) do
      v:kbd()
    end
  end
end

--/ mouse

function Element:mouseOver( x, y )
  if self.moColor ~= nil then self.drawColor = self.moColor end
end

function Element:mouseAway()
  redraw = 1
end

function Element:mouseUp( x, y )
end

function Element:mouseWheel(v)
--  if self.action ~= nil and type( self.param ) == "table" then
--  end
end


function Element:hitTest( x, y )
  local thisX, thisY, thisW, thisH =
    self.drawx or self.x, self.drawy or self.y,
    self.drawW or self.w or 0, self.drawH or self.h
  local xS, yS = x / drawScale, y / drawScale
  if self.visible ~= false then
    local inrect =
      xS >= thisX and yS >= thisY and xS < thisX + thisW and yS < thisY + thisH
    if self.children ~= nil and
      ( inrect == true or self.has_children_outside == 1 ) then
      for i, v in pairs( self.children ) do
        local s = v:hitTest( x, y )
        if s ~= nil then return s end
      end
    end
    if inrect and ( self.interractive ~= false ) then
      return self
    end
  end
  return nil
end

function Element:setDrawColor( colr )
  if colr ~= nil then
    self.drawColor = colr
  end
end

function Element:dropDrawColor()
  if self.drawColor ~= nil then
    self.drawColor = nil
  end
end



--/ Label
-- needs recosntuct to use instead of kDrawElement()
Label = Element:new()
function Label:new( parent, o )
  self.__index = self
  self.flow = true
  self.x, self.h, self.w, self.border =
    2, 50, o.w or 64, o.border or ''

  adoptChild( parent, o )
  setmetatable( o, self )
  return o
end

--/ mouse

function Label:mouseOver()
  self.drawColor = ( self.color or clr.default ) + 0x10101000
  root:doUpdateState()
--  redraw = 1
end

function Label:mouseAway()
  self.drawColor = nil
  root:doUpdateState()
  redraw = 1
end

function Label:mouseUp()
  if self.action ~= nil then self.action( self.param ) end
  root:doUpdateState()
--  redraw = 1
end


--/ ScrollBox
ScrollBox = Element:new()
function ScrollBox:new( parent, o )
  self.__index = self
  self.flow = true
  self.itemplate = o.itemplate
--  self.item = {}
  self.list = o.list
  self.x, self.y, self.h, self.w, self.drawH, self.border =
    2, o.y or 2, 256, o.w or 360, o.drawH or o.w, o.border or ''

  -- indicate current cursor position
  self.cursor = { pos = 1, prev = 1 }
  -- how many items are visible
  self.visItems = self.drawH // (self.itemplate.drawH+2)
  -- first visible item from 'list' table
  self.visListStart = 1
  -- last visble inside scrollbox viewport
  self.visListEnd = self.visListStart + self.visItems - 1
  
  self.pos = { x = 50, y = 0 }
  self.kspc = false -- space bar toggled? true|false
  self.tspc = 0
  self.playing = false

  self:upd()
  self:initems( self.visListStart, #self.list )

  adoptChild( parent, o )
  setmetatable( o, self )
  return o
end

function ScrollBox:initems( istart, iend )
  local st = self.itemplate
  if istart == nil then istart = self.visListStart end
  if iend == nil then iend = self.visListEnd end
  
  self.pos.dyStart = self.y + 2
  self.pos.dy = self.pos.dyStart
  self.pos.dyStep = st.drawH + 2
  for i = istart, iend do
    --self.item[ i ] = 
    Label:new( self, {
      nm = string.format( 'item%d', i ),
      visible = false,
      x = self.pos.x, y = self.pos.dy,
      drawW = st.drawW, drawH = st.drawH,
      color = st.color, action = st.action, param = st.param,
      text = { str = self.list[ i ].name, col = st.text.col,
      style = st.text.style, align = st.text.align }
    })
    self.pos.dy = self.pos.dyStart + self.pos.dyStep * i
  end
  self:updCursorItem()
  self:rebldVisList()

end

function ScrollBox:delitems()
  if self.children ~= nil then
    for i, v in ipairs( self.children ) do
      v.parent = nil
      v = nil
    end
    self.children = nil
  end
  --for i = self.visListStart, self.visItems do
  --  if self.item[ i ] ~= nil then
  --    self.item[ i ].parent = nil
  --    self.item[ i ] = nil
  --  end
  --end
end

function ScrollBox:kbd()
  -- update cursor
  -- position
  local kbdo = 0
  if s_char == kbd.down then
    if self.cursor.pos < #self.list then
      self.cursor.prev = self.cursor.pos
      self.cursor.pos = self.cursor.pos + 1
      kbdo = 1
    end
  end
  if s_char == kbd.up then
    if self.cursor.pos > 1 then
      self.cursor.prev = self.cursor.pos
      self.cursor.pos = self.cursor.pos - 1
      kbdo = 1
    end
  end
  if self.list ~= nil and self.list[ self.cursor.pos ] ~= nil then
    reaper.SetEditCurPos( self.list[ self.cursor.pos ].pos, true, true )
  end

  if s_char == kbd.space then
    local ispl = reaper.GetPlayState() -- &1=playing,&2=pause,&=4 is recording 
    if ispl == 0 or ispl == 2 then -- stopped or paused
        reaper.Main_OnCommand( 1007, 0 ) -- play
    else
        reaper.Main_OnCommand( 1016, 0 ) -- transport: stop
    end
  end
  
  -- chk bounds
  -- vis
  if kbdo == 1 then
    self:upd()
    self:updCursorItem()
    self:rebldVisList()
    root:doUpdateState()
    kbdo = false
    redraw = 1
  end
end


function ScrollBox:updCursorItem()
  local pos = self.cursor.pos 
  self.children[ pos ]:setDrawColor( self.itemplate.colcursor )
  self.children[ pos ].color = nil
  
  local prev = self.cursor.prev 
  if prev ~= pos and prev >= 1 then
    self.children[ prev ]:dropDrawColor()
    self.children[ prev ].color = self.itemplate.color
  end
end


function ScrollBox:rebldVisList()
  self.pos.dy = self.pos.dyStart
  local di = 1
  for i = self.visListStart, self.visListEnd do
    if self.children[ i ] == nil then break end

    if i > 1 and
      i == self.visListStart
    then self.children[ i-1 ].visible = false end
    if i < #self.list and
      i == self.visListEnd
    then self.children[ i+1 ].visible = false end

    -- visible only childs matched the parent box bounds
    if i >= self.visListStart and i <= self.visListEnd then
      self.children[ i ].visible = true
      self.children[ i ].y = self.pos.dy
      self.pos.dy = self.pos.dyStart + self.pos.dyStep * di
      di = di + 1
    end
  end
end


function ScrollBox:upd()
  -- chk is cursor out of visListItems chunk
  if self.cursor.pos > self.visListEnd and self.visListEnd < #self.list then
    -- slide list down by one
    self.visListStart = self.visListStart + 1
  end
  if self.cursor.pos < self.visListStart and self.visListStart > 1 then
    -- slide list up by one
    self.visListStart = self.visListStart - 1
  end

  -- chk visList var pair, limiting cursor bounds by list length
  if self.visListStart + (self.visItems-1) >= #self.list then
    self.visListEnd = #self.list
  else
    self.visListEnd = self.visListStart + (self.visItems - 1)
  end
  if self.visListEnd - (self.visItems-1) <= 1 then
    self.visListStart = 1
  else
    self.visListStart = self.visListEnd - (self.visItems - 1)
  end
--[[
  local msg = string.format(
    'scrl.upd| cpos: %d, cprev: %d, vitms: %s, vlst: %s, vle: %s \r\n',
    self.cursor.pos, self.cursor.prev, self.visItems, self.visListStart, self.visListEnd
  ) reaper.ShowConsoleMsg( msg ) 
]]--
end


--/ main loop
function rloop()
  -- kbd
  s_char = gfx.getchar()
  root:kbd()
  
  -- mouse stuff
  local isCap = ( gfx.mouse_cap & 1 )
  if gfx.mouse_x ~= mouseXold or
    gfx.mouse_y ~= mouseYold or
    isCap ~= mouseCapOld or
    gfx.mouse_wheel ~= 0 then
    --
    local wheel_amt = 0
    if gfx.mouse_wheel ~= 0 then
      mouseWheelAccum = mouseWheelAccum + gfx.mouse_wheel
      gfx.mouse_wheel = 0
      wheel_amt = math.floor( mouseWheelAccum / 120 + 0.5 )
      if wheel_amt ~= 0 then mouseWheelAccum = 0 end
    end
    
    local hit = nil --root:hitTest( gfx.mouse_x, gfx.mouse_y )
    if isCap == 0 and mouseCapOld == 1 then -- mouse up
      if activeMouseElement ~= nil and hit == activeMouseElement then -- still hover
        activeMouseElement:mouseUp( gfx.mouse_x, gfx.mouse_y )
      end
    end
    if isCap == 0 and mouseCapOld == 0 then -- uncaptured mouse dn or mv
      if activeMouseElement ~= nil and activeMouseElement ~= hit then
        activeMouseElement:mouseAway()
      end
      activeMouseElement = hit
    end
    
    if activeMouseElement ~= nil then
      if isCap == 0 or mouseCapOld == 0 then --uncaptured mouse dn or mv
        activeMouseElement:mouseOver()
      end
      if wheel_amt ~= 0 then
        activeMouseElement:mouseWheel( wheel_amt )
      end
    end
    mouseXold, mouseYold, mouseCapOld = gfx.mouse_x, gfx.mouse_y, isCap
  end
  
  if resize == 1 or
    root.drawW ~= gfx.w * drawScale or root.drawH ~= gfx.h * drawScale then
    -- wnd resized
    root.drawW, root.drawH = gfx.w * drawScale, gfx.y * drawScale
    root:onSize()
    root:draw()
    resize, redraw = 0, 0
  elseif redraw == 1 then
    root:draw()
    redraw = 0
  end

end

local function Main()
  rloop()
  gfx.update()
  if s_char >= 0 then
    if s_char ~= 27 and s_char ~= -1 then
      reaper.defer( Main )
    end
  end
end


---///---

  
--/ retrieve reaper's markers/regions
  local r, nMarkers, nRegions = reaper.CountProjectMarkers(0)
  tM = {}
  tR = {}

  local cr, cm = 1, 1
  for x = 0, nMarkers-1 do
    local r, isrgn, ps, rgnend, nm, mIdx = reaper.EnumProjectMarkers( x )
    if isrgn then
      tR[cr] = { idx = mIdx, name = nm, pos = ps, rend = rgnend }
      cr = cr + 1
    else
      tM[cm] = { idx = mIdx, name = nm, pos = ps }
      cm = cm + 1
    end
  end

--/ build elements hierarchy

root = Element:new( nil,
  {
    nm = 'root',
    x = 0, y = 0, drawx = 0, drawy = 0,
    drawW = guirect.br.x, drawH = guirect.br.y,
    color = 0x2c0c00ff -- 0x4c1c0cff - dark orange / brown
  } )
  

scrl = ScrollBox:new( root, {
    nm = 'scrl',
    --crop = true,
    drawW = 630, drawH = 366,
    x = 48, y = 100,
    color = 0x204000ff,
    colcursor = 0x201000ff,
    list = tM,
    itemplate = {
      w = 620, drawW = 620, h = 50, drawH = 50,
      color = 0x402000ff, param = nil,
      text = { col = clr.txt.default, style = 1, align = 5 }
    }
  })

---///---


gfx.init( "Big Markers List", guirect.br.x, guirect.br.y, 0, 1024, 50)
Main()
redraw = 1

function Quit()
  gfx.quit();
end

reaper.atexit( Quit )


